
/*
  Released under the antivirial license.  Basically, you can do anything
  you want with it as long as what you want doesn't involve the GNU GPL.
  See http://www.ecstaticlyrics.com/antiviral/ for more information.
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <GL/glfw.h>
#include <GL/glu.h>
#include "stb_image.h"
#include "sha1.h"

struct int_xy {
  int x;
  int y;
};

struct item { // A file or a folder...
  char *name;
  char *pathname;
  long long size;
  struct int_xy one;
  struct int_xy two;
  struct item *parent;
  int index;
  struct item **contents;
  float r, g, b;
};

int filesystem;
struct item *current_item;
struct item *selected_item;
int window_width;
int window_height;

GLint font_texture = 0;

GLuint load_image(char *filename) {
  int width, height;
  char *image = stbi_load(filename, &width, &height, NULL, 4);
  if (image == NULL) {
    fprintf(stderr, "Failed to load %s\n", filename);
    exit(1);
  };
  GLuint texture;
  glGenTextures(1, &texture);
  glBindTexture(GL_TEXTURE_2D, texture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, image);
  stbi_image_free(image);
  return texture;
};

static void sort_items(struct item **array, int count) {
  if (count <= 1) return;
  int half_one = count >> 1;
  int half_two = count - half_one;
  sort_items(array, half_one);
  sort_items(array + half_one, half_two);
  struct item **temp = malloc(sizeof(void *) * count);
  struct item **pointer_one = array;
  struct item **pointer_two = array + half_one;
  for (int i = 0; i < count; i++) {
    if (half_one && half_two) {
      if ((*pointer_one)->size >= (*pointer_two)->size) {
        temp[i] = *pointer_one;
        pointer_one++;
        half_one--;
      } else {
        temp[i] = *pointer_two;
        pointer_two++;
        half_two--;
      };
    } else if (half_one) {
      temp[i] = *pointer_one;
      pointer_one++;
      half_one--;
    } else {
      temp[i] = *pointer_two;
      pointer_two++;
      half_two--;
    };
  };
  memmove(array, temp, sizeof(void *) * count);
  free(temp);
};

int mouse_x = 0;
int mouse_y = 0;

void GLFWCALL position_callback(int x, int y) {
  mouse_x = x;
  mouse_y = window_height - y - 1;
};

void GLFWCALL button_callback(int button, int action) {
  if (action != GLFW_PRESS) return;
  if (button == 0 || button == 1) {
    if (selected_item != NULL && selected_item->parent != NULL) {
      if (button == 0) printf("%s\n", selected_item->pathname);
      long long size = selected_item->size;
      struct item *parent = selected_item->parent;
      for (int i = 0; i < parent->index; i++) {
        if (parent->contents[i] == selected_item) {
          memmove(parent->contents + i, parent->contents + i + 1, (parent->index - i - 1) * sizeof(void *));
          parent->index--;
          break;
        };
      };
      while (1) {
        parent->size -= size;
        parent = parent->parent;
        if (parent == NULL) break;
        sort_items(parent->contents, parent->index);
      };
    };
  };
  if (button == 2 && selected_item != NULL) {
    current_item = selected_item;
  };
};

void GLFWCALL wheel_callback(int position) {
  static int memory = 0;
  for (; memory < position; memory++) {
    if (selected_item != NULL) {
      struct item *target = selected_item;
      while (target->parent != NULL && target->parent != current_item) {
        target = target->parent;
      };
      if (target->parent != NULL) current_item = target;
    };
  };
  for (; memory > position; memory--) {
    if (current_item->parent != NULL) {
      current_item = current_item->parent;
    };
  };
};

void GLFWCALL key_callback(int key, int action) {
  if (action != GLFW_PRESS) return;
  if (key == GLFW_KEY_ESC) exit(0);
};

void scan_folder(struct item *folder) {
  #define folder (*folder)

  printf("Scanning folder '%s'\n", folder.pathname);

  DIR *dir;
  dir = opendir(folder.pathname);
  if (dir == NULL) {
    fprintf(stderr, "Error %d while opening '%s'\n", errno, folder.pathname);
    exit(1);
  };
  struct dirent *entry;
  while (entry = readdir(dir)) {
    if (entry->d_name[0] == '.') {
      if (entry->d_name[1] == 0) continue;
      if (entry->d_name[1] == '.') {
        if (entry->d_name[2] == 0) continue;
      };
    };
    folder.contents = realloc(folder.contents, (folder.index + 1) * sizeof(void *));
    folder.contents[folder.index] = malloc(sizeof(struct item));
    memset(folder.contents[folder.index], 0, sizeof(struct item));
    #define that (*folder.contents[folder.index])
    int length = strlen(entry->d_name);
    that.name = malloc(length + 1);
    strcpy(that.name, entry->d_name);
    if (folder.pathname[strlen(folder.pathname) - 1] == '/') {
      that.pathname = malloc(strlen(folder.pathname) + length + 1);
      sprintf(that.pathname, "%s%s", folder.pathname, that.name);
    } else {
      that.pathname = malloc(strlen(folder.pathname) + 1 + length + 1);
      sprintf(that.pathname, "%s/%s", folder.pathname, that.name);
    };
    that.parent = &folder;
    struct stat buf;
    if (lstat(that.pathname, &buf) == -1) {
      fprintf(stderr, "Error %d while lstat()ing '%s'\n", errno, that.pathname);
      exit(1);
    };
    if (buf.st_dev == filesystem) {
      if (S_ISLNK(buf.st_mode)) {
      };
      if (S_ISDIR(buf.st_mode)) {
        scan_folder(&that);
      };
      if (S_ISREG(buf.st_mode)) {
        struct item *current = &that;
        while (current != NULL) {
          current->size += buf.st_size;
          current = current->parent;
        };
      };
    };
    // Assign a color based on hash of pathname...
    char sum[20];
    sha1(sum, that.pathname, strlen(that.pathname));
    that.r = 0.6 * sum[0] / 255.0 + 0.2;
    that.g = 0.6 * sum[1] / 255.0 + 0.2;
    that.b = 0.6 * sum[2] / 255.0 + 0.2;
    #undef that
    folder.index++;
  };
  closedir(dir);
  sort_items(folder.contents, folder.index);
  #undef folder
};

void draw_text(int x, int y, char *text) {
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND); glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glColor4f(1.0, 1.0, 1.0, 1.0);
  glBindTexture(GL_TEXTURE_2D, font_texture);
  glBegin(GL_QUADS);
  for (int i = 0; i < strlen(text); i++) {
    int c = text[i];
    if (c < 32 || c > 126) c = 127;
    int cx = c % 16;
    int cy = c / 16;
    glTexCoord2d((cx + 0) / 16.0, (cy + 0) / -8.0); glVertex2f(8 * i + x, y);
    glTexCoord2d((cx + 1) / 16.0, (cy + 0) / -8.0); glVertex2f(8 * i + x + 8, y);
    glTexCoord2d((cx + 1) / 16.0, (cy + 1) / -8.0); glVertex2f(8 * i + x + 8, y + 16);
    glTexCoord2d((cx + 0) / 16.0, (cy + 1) / -8.0); glVertex2f(8 * i + x, y + 16);
  };
  glEnd();
  glDisable(GL_TEXTURE_2D);
};

void draw_folder (struct item *folder, int nest) {
  #define folder (*folder)
  struct int_xy one = folder.one;
  struct int_xy two = folder.two;

  int width = two.x - one.x;
  int height = two.y - one.y;

  one.x++; two.x--; width -= 2;
  one.y++; two.y--; height -= 2;

  if (width < 1 || height < 1) return;

  if (mouse_x >= one.x && mouse_x < two.x) {
    if (mouse_y >= one.y && mouse_y < two.y) {
      selected_item = &folder;
    };
  };

  glColor3f(folder.r, folder.g, folder.b);
  glBegin(GL_QUADS);
  glVertex2f(one.x, one.y);
  glVertex2f(two.x, one.y);
  glVertex2f(two.x, two.y);
  glVertex2f(one.x, two.y);
  glEnd();

  one.x += 3; one.y += 3;
  two.x -= 3; two.y -= 3;
  width -= 6; height -= 6;

  if (width >= 19 && height >= 19) {

    char size_text[64];
    if (folder.size > 1000000000) {
      snprintf(size_text, 64, "%.3g GB", folder.size / 1000000000.0);
    } else if (folder.size > 1000000) {
      snprintf(size_text, 64, "%.3g MB", folder.size / 1000000.0);
    } else if (folder.size > 1000) {
      snprintf(size_text, 64, "%.3g kB", folder.size / 1000.0);
    } else {
      snprintf(size_text, 64, "%.3g B", folder.size / 1.0);
    };
    size_text[63] = 0;
    int length = strlen(size_text);

    if (width >= height) {
      int max = width / 8;
      if (strlen(folder.name) <= max) {
        draw_text(one.x, two.y - 16, folder.name);
      } else {
        char name[max + 1];
        memmove(name, folder.name, max);
        name[max] = 0;
        name[max-1] = 127;
        draw_text(one.x, two.y - 16, name);
      };
      if (two.x - one.x > 8 * (strlen(folder.name) + length + 2)) {
        draw_text(two.x - length * 8, two.y - 16, size_text);
      };
      two.y -= 19;
      height -= 19;
    } else {
      glPushMatrix();
      glTranslated(one.x + 16, one.y, 0);
      glRotated(90, 0, 0, 1);
      int max = height / 8;
      if (strlen(folder.name) <= max) {
        draw_text(0, 0, folder.name);
      } else {
        char name[max + 1];
        memmove(name, folder.name, max);
        name[max] = 0;
        name[max-1] = 127;
        draw_text(0, 0, name);
      };
      if (two.y - one.y > 8 * (strlen(folder.name) + length + 2)) {
        draw_text(two.y - one.y - length * 8, 0, size_text);
      };
      glPopMatrix();
      one.x += 19;
      width -= 19;
    };

  };

  int first = 0;
  int consumed_y = 0;
  int remaining_y = height;
  long long remaining_size = folder.size;

  // Determine the best number of items to show on each row within the
  // folder such that the ratio between width and height is most equal.

  // We do this by first assuming we'll put every item in a single row,
  // then we pop items off the end until the situation ceases to improve.

  while (first < folder.index) {

    if (width < 3 || height - consumed_y < 3) return;

    int last = folder.index - 1;
    long long row_size = remaining_size;

    double previous_worst_ratio = 0.0;

    long long selected_row_size = row_size;
    int selected_last = last;

    while (1) {
      double estimated_y = (double) remaining_y * row_size / remaining_size;
      long long first_size = folder.contents[first]->size;
      long long last_size = folder.contents[last]->size;
      double first_width = (double) width * first_size / row_size;
      double last_width = (double) width * last_size / row_size;
      double first_ratio, last_ratio;
      if (first_width < estimated_y) {
        first_ratio = (double) first_width / estimated_y;
      } else {
        first_ratio = (double) estimated_y / first_width;
      };
      if (last_width < estimated_y) {
        last_ratio = (double) last_width / estimated_y;
      } else {
        last_ratio = (double) estimated_y / last_width;
      };
      double worst_ratio;
      if (first_ratio < last_ratio) {
        worst_ratio = first_ratio;
      } else {
        worst_ratio = last_ratio;
      };
      if (worst_ratio < previous_worst_ratio) break;
      previous_worst_ratio = worst_ratio;
      selected_row_size = row_size;
      selected_last = last;
      if (first == last) break;
      row_size -= folder.contents[last]->size;
      last--;
    };

    row_size = selected_row_size;
    last = selected_last;

    int actual_y = round((double) remaining_y * row_size / remaining_size);
    if (actual_y < 3) break;

    long long sum = 0;
    for (int i = first; i <= last; i++) {
      int sx = round((double) width * sum / row_size);
      sum += folder.contents[i]->size;
      int tx = round((double) width * sum / row_size);
      folder.contents[i]->one.x = one.x + sx;
      folder.contents[i]->two.x = one.x + tx;
      folder.contents[i]->one.y = one.y + consumed_y;
      folder.contents[i]->two.y = one.y + consumed_y + actual_y;
      if (nest < 2) {
//        printf("Draw '%s' at %d, %d, %d, %d\n", folder.contents[i]->name, folder.contents[i]->one.x, folder.contents[i]->two.x, folder.contents[i]->one.y, folder.contents[i]->two.y);
      };
      draw_folder(folder.contents[i], nest + 1);
    };

    remaining_size -= row_size;
    consumed_y += actual_y;
    remaining_y -= actual_y;
    first = last + 1;

  };

  #undef folder
};

int main(int argc, char **argv) {

  struct item root = {};
  root.parent = NULL;

  if (argc >= 2) {
    root.name = argv[1];
  } else {
    root.name = "/";
  };

  root.pathname = root.name;

  struct stat buf;
  if (lstat(root.pathname, &buf) == -1) {
    fprintf(stderr, "Error %d while lstat()ing '%s'\n", errno, root.pathname);
    exit(1);
  };
  filesystem = buf.st_dev;

  scan_folder(&root);

  current_item = &root;

  glfwInit();
  glfwOpenWindow(960, 720, 8, 8, 8, 8, 0, 0, GLFW_WINDOW);
  glfwSetWindowTitle("Pj's File Size Viewer");
  glfwSetMousePosCallback(position_callback);
  glfwSetMouseButtonCallback(button_callback);
  glfwSetMouseWheelCallback(wheel_callback);
  glfwSetKeyCallback(key_callback);

  font_texture = load_image("font.png");

  while (glfwGetWindowParam(GLFW_OPENED)) {

    glfwGetWindowSize(&window_width, &window_height);
    glViewport(0, 0, window_width, window_height);
    glMatrixMode(GL_MODELVIEW); glLoadIdentity();
    glTranslated(-1.0, -1.0, 0.0);
    glScaled(2.0 / window_width, 2.0 / window_height, 1.0);

    current_item->one.x = 0;
    current_item->one.y = 0;
    current_item->two.x = window_width;
    current_item->two.y = window_height;

    static struct item *last_selected = NULL;
    selected_item = NULL;

    glClearColor(0.0, 0.0, 0.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    draw_folder(current_item, 0);

    if (selected_item != last_selected) {
      if (selected_item) {
        glfwSetWindowTitle(selected_item->pathname);
      } else {
        glfwSetWindowTitle("Pj's File Size Viewer");
      };
      last_selected = selected_item;
    };

    glfwSwapBuffers();
    glfwSleep(0.01);

  };

  glfwCloseWindow();

};
